// Ryan Schmidt
// Copyright (c) 2006-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt


#include "stdafx.h"
#include "glut.h"
#include "MeshObject.h"
#include <ExtendedWmlCamera.h>

MeshObject g_mesh;
rms::VFTriangleMesh g_initial_mesh;
int g_normal_smooth_iters = 0;
bool g_draw_normals = false;
bool g_draw_wireframe = false;
bool g_draw_help = false;
Wml::ExtendedWmlCamera * g_pCamera = NULL;

// input modes for mouse handler
enum INPUT_MODE {
	NONE,
	DOLLY,
	ORBIT,
	PAN,
	MOVE_EXPMAP_SEED,
	MOVE_EXPMAP_SEED2,
	SCALE_EXPMAP,
	ROTATE_EXPMAP
};
INPUT_MODE g_inputMode = NONE;


// display layout modes
enum DISPLAY_MODE {
	MESH,
	MESH_AND_QUARTERUV,
	UVONLY,

	LAST_MODE		// leave at the end!
};
DISPLAY_MODE g_displayMode = MESH_AND_QUARTERUV;


// input device state
bool g_bAltButtonDown = false;
bool g_bCtrlButtonDown = false;
bool g_bShiftButtonDown = false;
bool g_bLeftMouseDown = false;
bool g_bMiddleMouseDown = false;
bool g_bRightMouseDown = false;
Wml::Vector2f g_vLastMousePos = Wml::Vector2f::ZERO; 


void drawString(const char * s)
{
  unsigned int i;
  for (i = 0; i < strlen (s); i++)
    glutBitmapCharacter (GLUT_BITMAP_8_BY_13, s[i]);
}
void drawString (const std::string & s)
{
	drawString(s.c_str());
}


void drawHelpText()
{
	GLint nViewport[4];
	glGetIntegerv(GL_VIEWPORT, nViewport);

	glPushAttrib(GL_ENABLE_BIT);

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glColor3f(1,0,0);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(nViewport[0], nViewport[0]+nViewport[2], nViewport[1], nViewport[1]+nViewport[3], -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	const char * pHelp[] = {
		"ALT+LEFT/MIDDLE/RIGHT to rotate/pan/zoom",
		"LEFT click/drag to move expmap center",
		"CTRL+LEFT to scale expmap, CTRL+RIGHT to rotate",
		"[f] to load new file",
		"[d] to cycle display modes",
		"[t] to toggle display normals",
		"[w] to toggle display wireframe",
		"[q] to quit"
	};
	int nHelp = 8;

	int nStep = 10;
	int nStart = nViewport[1]+1;
	if ( g_draw_help ) {
		nStep += 2;
		nStart += nStep*nHelp;
	}
	glRasterPos2f( nViewport[0]+10, nStart+5*nStep );
	drawString( _RMSInfoString("[u] Upwind averaging: %s", (g_mesh.GetUpwindAverage() ? "On" : "Off") ) );
	glRasterPos2f( nViewport[0]+10, nStart+4*nStep );
	drawString( _RMSInfoString("[n] Normal smoothing: %s", (g_mesh.GetSmoothNormals() ? "On" : "Off") ) );
	glRasterPos2f( nViewport[0]+10, nStart+3*nStep );
	drawString( _RMSInfoString("[c] Constant normal:  %s", (g_mesh.GetUseConstantNormal() ? "On" : "Off") ) );
	glRasterPos2f( nViewport[0]+10, nStart+2*nStep );
	drawString( _RMSInfoString("[l] Preserve lengths: %s", (g_mesh.GetPreserveProjectedLengths() ? "On" : "Off") ) );
	glRasterPos2f( nViewport[0]+10, nStart+nStep );
	drawString( _RMSInfoString("Normal smooth iters:%3d %s", g_normal_smooth_iters, 
		(g_draw_help) ? "([s] +1  [S] +25)" : "" ) );


	if ( g_draw_help ) {
		for ( int k = 0; k < nHelp; ++k ) {
			glRasterPos2f( nViewport[0]+10, nViewport[1]+(k+1)*nStep );
			drawString(pHelp[(nHelp-k-1)]);
		}
		glRasterPos2f( nViewport[0]+10, nViewport[1] );
		drawString( "Press h for less help" );
	} else {
		glRasterPos2f( nViewport[0]+10, nViewport[1] );
		drawString( "Press h for more help" );
	}


	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);


	glPopAttrib();
}


// this is (well, should be) the only platform-dependent bit of code...
void LoadNewMesh(const char * pFilename = NULL)
{
	std::string sFilename;
	if ( pFilename == NULL ) {
		CFileDialog dialog(TRUE, "obj", "mesh.obj", 0, "OBJ Files (*.obj)|*.obj|All Files (*.*)|*.*", NULL);
		if ( dialog.DoModal() != IDOK ) 
			return;
		sFilename = std::string(dialog.GetPathName());
	} else
		sFilename = std::string(pFilename);

	g_mesh.ReadMeshOBJ(sFilename.c_str());
	g_initial_mesh.Copy(g_mesh.GetMesh());
	g_normal_smooth_iters = 0;
}



void glut_initgl()
{
	g_pCamera = new Wml::ExtendedWmlCamera(1.0f,1.0f);
	// set default frame (lookat)
	g_pCamera->SetTargetFrame( Wml::Vector3f(0.0, 0.0, 5.0f), 
		Wml::Vector3f(-1.0f, 0.0, 0.0),
		Wml::Vector3f(0.0, 1.0f, 0.0), 
		Wml::Vector3f(0.0, 0.0, 0.0));

	// enable color material
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

	// initialize depth test
	glEnable(GL_DEPTH_TEST);

	// enable lighting
	glEnable(GL_LIGHTING);

	// set ambient light
	GLfloat ambient[] = { 0.1f, 0.1f, 0.1f, 1.0f };
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);

	// enable 2 lights
	glEnable(GL_LIGHT0);
	float fEye = (g_pCamera->GetLocation() - g_pCamera->GetTarget()).Length();
	float vUpLeft[4] = {-2.0f*fEye, fEye, fEye, 1.0f};
	glLightfv(GL_LIGHT0, GL_POSITION, vUpLeft);
	float fAmbient1[4] = {0,0,0,1};
	glLightfv(GL_LIGHT0, GL_AMBIENT, fAmbient1);
	float fDiffuse1[4] = {0.6f, 0.6f, 0.6f, 1.0f};
	glLightfv(GL_LIGHT0, GL_DIFFUSE, fDiffuse1);
	float fSpec1[4] = {0.5f, 0.5f, 0.5f,1.0f};
	glLightfv(GL_LIGHT0, GL_SPECULAR, fSpec1);

	glEnable(GL_LIGHT1);
	float vUpRight[4] = { 2.0f*fEye, fEye, fEye, 1.0f};
	glLightfv(GL_LIGHT1, GL_POSITION, vUpRight);
	float fAmbient2[4] = {0,0,0,1};
	glLightfv(GL_LIGHT1, GL_AMBIENT, fAmbient2);
	float fDiffuse2[4] = {0.5f, 0.5f, 0.5f, 1.0f};
	glLightfv(GL_LIGHT1, GL_DIFFUSE, fDiffuse2);
	float fSpec2[4] = {0.5f, 0.5f, 0.5f,1.0f};
	glLightfv(GL_LIGHT1, GL_SPECULAR, fSpec2);

	// set fixed specular material color
	float fSpecular[] = {1.0f, 1.0f, 1.0f, 1.0f};
	glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, fSpecular );
	glMaterialf( GL_FRONT_AND_BACK, GL_SHININESS, 40.0f );
}



void glut_main_display(void)
{
	// make sure expmap is computed
	g_mesh.ValidateExpMap();

	// clear window
	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


	// render the 3D view
	if ( g_displayMode == MESH || g_displayMode == MESH_AND_QUARTERUV ) {
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		g_pCamera->Update();
		glEnable(GL_LIGHTING);

		g_mesh.Render(g_draw_wireframe, g_draw_normals);

		//glPushAttrib(GL_ENABLE_BIT | GL_POINT_BIT);
		//glDisable(GL_LIGHTING);
		//glColor3f(0.0f,1.0f,0.0f);
		//glPointSize(10.0f);
		//glBegin(GL_POINTS);
		//glVertex3fv( g_mesh.ExpMapFrame().Origin() );
		//glEnd();
		//glPopAttrib();
	}


	if ( g_displayMode == MESH_AND_QUARTERUV || g_displayMode == UVONLY ) {

		// render the 2D overlay
		const unsigned int * viewport = g_pCamera->GetViewport();
		int nDivider = (g_displayMode == MESH_AND_QUARTERUV) ? 2 : 1;
		int nMinDim = std::min(viewport[2], viewport[3]);
		//glViewport(0, 0, nMinDim/nDivider, nMinDim/nDivider);
		glViewport(viewport[2]-nMinDim/nDivider, 0, nMinDim/nDivider, nMinDim/nDivider);
		
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		gluOrtho2D(-1, 1, -1, 1);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		g_mesh.RenderParamMesh();

		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glViewport(viewport[0],viewport[1],viewport[2],viewport[3]);
	}

	drawHelpText();

	glutSwapBuffers();
};




void glut_main_reshape (int cx, int cy)
{
	// set viewport and frustum, and update camera
	g_pCamera->SetViewPort(0, 0, cx, cy);
	g_pCamera->SetFrustum( 30.0f, float(cx)/float(cy), 0.1f, 100.0);
	g_pCamera->Update();
};


void glut_mouse(int button, int state, int x, int y)
{
	g_bAltButtonDown = ( glutGetModifiers() == GLUT_ACTIVE_ALT );
	g_bCtrlButtonDown = ( glutGetModifiers() == GLUT_ACTIVE_CTRL );
	g_bShiftButtonDown = ( glutGetModifiers() == GLUT_ACTIVE_SHIFT );
	g_vLastMousePos = Wml::Vector2f(x,y);

	g_bRightMouseDown = g_bMiddleMouseDown = g_bLeftMouseDown = false;
	
	if ( button == GLUT_RIGHT_BUTTON ) {
		g_bRightMouseDown = ( state == GLUT_DOWN ) ? true : false;
	} else if ( button == GLUT_LEFT_BUTTON ) {
		g_bLeftMouseDown = ( state == GLUT_DOWN ) ? true : false;
	} else if ( button == GLUT_MIDDLE_BUTTON ) {
		g_bMiddleMouseDown = ( state == GLUT_DOWN ) ? true : false;
	}

	g_inputMode = NONE;
	if ( g_bAltButtonDown && g_bRightMouseDown ) 
		g_inputMode = DOLLY;
	else if ( g_bAltButtonDown && g_bLeftMouseDown ) 
		g_inputMode = ORBIT;
	else if ( g_bAltButtonDown && g_bMiddleMouseDown ) 
		g_inputMode = PAN;
	else if ( g_bLeftMouseDown && g_bCtrlButtonDown )
		g_inputMode = SCALE_EXPMAP;
	else if ( g_bRightMouseDown && g_bCtrlButtonDown )
		g_inputMode = ROTATE_EXPMAP;
	//else if ( g_bLeftMouseDown && g_bShiftButtonDown )
	//	g_inputMode = MOVE_EXPMAP_SEED2;
	else if ( g_bLeftMouseDown )
		g_inputMode = MOVE_EXPMAP_SEED;
}



void glut_mouse_motion(int x, int y)
{
	static const float fPanScale = 0.01f;
	static const float fZoomScale = 0.01f;
	static const float fOrbitScale = 0.01f;
	static const float fExpMapScale = 0.005f;

	Wml::Vector2f vCur(x,y);
	Wml::Vector2f vDelta(vCur - g_vLastMousePos);
	g_vLastMousePos = vCur;

	switch ( g_inputMode ) {
		case DOLLY:
			g_pCamera->DollyZoom((-vDelta.Y() + vDelta.X()) * fZoomScale);
			glutPostRedisplay();
			break;
		case PAN:
			g_pCamera->PanLateral(vDelta.X() * fPanScale);
			g_pCamera->PanVertical(vDelta.Y() * fPanScale);
			glutPostRedisplay();
			break;
		case ORBIT:
			g_pCamera->OrbitLateral(-vDelta.X() * fOrbitScale);
			g_pCamera->OrbitVertical(vDelta.Y() * fOrbitScale);
			glutPostRedisplay();
			break;

		case MOVE_EXPMAP_SEED:
			{
				const unsigned int * pViewport = g_pCamera->GetViewport();
				Wml::Ray3f ray;
				g_pCamera->GetPickRay( vCur.X(), vCur.Y(), 
					pViewport[2], pViewport[3], ray.Origin, ray.Direction );
				g_mesh.MoveExpMap(ray);
				//g_mesh.AlignSeedFrame2(g_pCamera);
			}
			glutPostRedisplay();
			break;


		case MOVE_EXPMAP_SEED2:
			{
				const unsigned int * pViewport = g_pCamera->GetViewport();
				Wml::Ray3f ray;
				g_pCamera->GetPickRay( vCur.X(), vCur.Y(), 
					pViewport[2], pViewport[3], ray.Origin, ray.Direction );
				g_mesh.MoveExpMap2(ray);
				//g_mesh.AlignSeedFrame2(g_pCamera);
			}
			glutPostRedisplay();
			break;

		case SCALE_EXPMAP:
			g_mesh.ScaleExpMap( 1.0f + (vDelta.X() - vDelta.Y())*fExpMapScale );
			glutPostRedisplay();
			break;

		case ROTATE_EXPMAP:
			g_mesh.RotateExpMap( -(vDelta.X() + vDelta.Y())*fExpMapScale );
			//g_mesh.AlignSeedFrame2(g_pCamera);
			glutPostRedisplay();
			break;
	}

}



// keyboard event handler
void glut_keyboard (unsigned char key, int x, int y)
{
	static int info_banner = 1;

	switch (key) {
	case 'f':	case 'F':
		LoadNewMesh();
		glutPostRedisplay();
		break;
	case 'h':
		g_draw_help = ! g_draw_help;
		glutPostRedisplay();
		break;


	case 'd':  case 'D':
		g_displayMode = (DISPLAY_MODE)((g_displayMode + 1) % LAST_MODE);
		glutPostRedisplay();
		break;

	case 'n':  case 'N':
		g_mesh.SetSmoothNormals( ! g_mesh.GetSmoothNormals() );
		_RMSInfo("Inline Normal smoothing is %s\n", (g_mesh.GetSmoothNormals() ? "Enabled" : "Disabled") );
		glutPostRedisplay();
		break;

	case 'u':  case 'U':
		g_mesh.SetUpwindAverage( ! g_mesh.GetUpwindAverage() );
		_RMSInfo("Upwind averaging is %s\n", (g_mesh.GetUpwindAverage() ? "Enabled" : "Disabled") );
		glutPostRedisplay();
		break;

	case 'c':
		g_mesh.SetUseConstantNormal( ! g_mesh.GetUseConstantNormal() );
		_RMSInfo("Constant normal is %s\n", (g_mesh.GetUseConstantNormal() ? "Enabled" : "Disabled") );
		glutPostRedisplay();
		break;

	case 'l':
		g_mesh.SetPreserveProjectedLengths( ! g_mesh.GetPreserveProjectedLengths() );
		_RMSInfo("Preserve projected lengths is %s\n", (g_mesh.GetPreserveProjectedLengths() ? "Enabled" : "Disabled") );
		glutPostRedisplay();
		break;

	case 's': 
		g_mesh.SmoothNormals();
		++g_normal_smooth_iters;
		glutPostRedisplay();
		break;
	case 'S':
		g_mesh.SmoothNormals(25);
		g_normal_smooth_iters += 25;
		glutPostRedisplay();
		break;

	case 'r':
		g_mesh.SetMesh(g_initial_mesh);
		g_normal_smooth_iters = 0;
		glutPostRedisplay();
		break;


	case 't':
		g_draw_normals = ! g_draw_normals;
		glutPostRedisplay();
		break;
	case 'w':
		g_draw_wireframe = ! g_draw_wireframe;
		glutPostRedisplay();
		break;


	case 'q':
	case 'Q':
		exit (0);
		break;
	};
};




void glut_idle (void)
{
};




CWinApp theApp;

int _tmain(int argc, _TCHAR* argv[])
{
	// init some MFC stuff so that CFileDialog won't throw any exceptions...
	InitCommonControls();
	AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0);

//	LoadNewMesh("sphere.obj");
	LoadNewMesh("bunny.obj");

	/* Glut initializations */
	glutInit (&argc, argv);
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	glutInitWindowPosition(5, 5);
	glutInitWindowSize(800, 600);

	/* Main window creation and setup */
	glutCreateWindow("Robust Discrete Exponential Map Demo");
	glutDisplayFunc(glut_main_display);
	glutReshapeFunc(glut_main_reshape);
	glutKeyboardFunc(glut_keyboard);
	glutMouseFunc(glut_mouse);
	glutMotionFunc(glut_mouse_motion);
	glutIdleFunc(glut_idle);

	glut_initgl();

	glutMainLoop();

	return 0;
}

