// Ryan Schmidt
// Copyright (c) 2006-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt

#include "StdAfx.h"
#include ".\meshobject.h"

#include <gl/GL.h>
#include <gl/GLU.h>
#include <VFMeshRenderer.h>
#include <string>
#include <MeshUtils.h>
#include <ExtendedWmlCamera.h>

MeshObject::MeshObject(void)
{
	m_bExpMapInitialized = false;
	m_bExpMap2Initialized = false;
	m_bExpMapValid = true;
	m_fDecalRadius = 0.3f;

	// create checker texture...
	std::vector<unsigned char> vTexBuf;
	Texture::MakeCheckerImage(512, 4, vTexBuf, true);
	m_texture.InitializeBuffer(512, 512, vTexBuf, GL_RGBA);
	m_texture.SetTextureInfo( GL_TEXTURE_2D );
	m_texture.TextureEnvMode() = GL_DECAL;
	m_texture.SetEdgeAlpha( 0 );

	m_bSmoothNormals = true;
	m_bUpwindAverage = true;
	m_bUseConstantNormal = false;
	m_bPreserveProjectedLengths = true;
}

MeshObject::~MeshObject(void)
{
}

void MeshObject::ReadMeshOBJ( const char * pFilename )
{
	rms::VFTriangleMesh tmp;
	std::string err;
	//bool bOK = m_mesh.ReadOBJ(pFilename,err);
	bool bOK = tmp.ReadOBJ(pFilename,err);
	if ( ! bOK ) {
		char buf[1024];
		sprintf_s(buf,1022,"Error opening file %s:\n%s\n", pFilename, err.c_str() );  buf[1023] = '\0';
		MessageBox(NULL, buf, "Error!", MB_OK);
		return;
	}

	m_bExpMapInitialized = false;
	m_bExpMap2Initialized = false;
	SetMesh(tmp);
}

void MeshObject::SetMesh( rms::VFTriangleMesh & mesh )
{
	m_mesh.Copy(mesh);
	m_bvTree.SetMesh(&m_mesh);
	m_expmapgen.SetSurface(&m_mesh, &m_bvTree);

	m_fMaxEdgeLength = m_mesh.GetMaxEdgeLength();

	if ( ! m_mesh.HasUVSet(0) )
		m_mesh.AppendUVSet();
	m_mesh.InitializeUVSet(0);

	m_fBoundaryWidth =  m_mesh.GetMaxEdgeLength() * 1.0f;

	m_mesh2.Copy(m_mesh);
	if ( ! m_mesh2.HasUVSet(0) )
		m_mesh2.AppendUVSet();
	m_mesh2.InitializeUVSet(0);


	if (m_bExpMapInitialized)
		m_bExpMapValid = false;
}


bool MeshObject::FindIntersection( Wml::Ray3f & vRay, Wml::Vector3f & vHit, Wml::Vector3f & vHitNormal )
{
	rms::IMesh::TriangleID tID;
	if ( ! m_bvTree.FindRayIntersection( vRay.Origin, vRay.Direction, vHit, tID ) )
		return false;
	Wml::Vector3f vVerts[3], vNorms[3];
	m_mesh.GetTriangle(tID, vVerts, vNorms);

	float fBary[3];
	rms::BarycentricCoords( vVerts[0], vVerts[1], vVerts[2], vHit, fBary[0], fBary[1], fBary[2] );
	vHitNormal = 
		fBary[0]*vNorms[0] + fBary[1]*vNorms[1] + fBary[2]*vNorms[2];
	vHitNormal.Normalize();

	return true;
}


void MeshObject::MoveExpMap( Wml::Ray3f & vRay )
{
	Wml::Vector3f vHit, vHitNormal;
	if ( ! FindIntersection(vRay, vHit, vHitNormal ) ) 
		return;
	m_vSeedFrame.Origin() = vHit;
	m_vSeedFrame.AlignZAxis(vHitNormal);
	m_bExpMapValid = false;
	m_bExpMapInitialized = true;
}

void MeshObject::MoveExpMap2( Wml::Ray3f & vRay )
{
	Wml::Vector3f vHit, vHitNormal;
	if ( ! FindIntersection(vRay, vHit, vHitNormal ) ) 
		return;
	m_vSeedFrame2.Origin() = vHit;
	m_vSeedFrame2.AlignZAxis(vHitNormal);
	m_bExpMapValid = false;
	m_bExpMap2Initialized = true;
}



float computeAlignAngle( const rms::Frame3f & withFrame, const rms::Frame3f & alignFrame,
						 Wml::Vector2f vFrameDir,
						 int viewport[4], double projection[16], double modelview[16], Wml::ExtendedWmlCamera * pCamera )
{
	float fScale = 0.25f;

	// project vFrameDir in withFrame to 2D screen-space vector
	Wml::Vector2d seedCenter, seedPos;   double winZ;
	Wml::Vector3f vOrigin = withFrame.Origin();
	gluProject( vOrigin.X(), vOrigin.Y(), vOrigin.Z(), modelview, projection, viewport, & seedCenter[0], &seedCenter[1], &winZ );
	Wml::Vector3f vPlanePos = vFrameDir.X()*withFrame.X() + vFrameDir.Y()*withFrame.Y();
	Wml::Vector3f vRightPos = vOrigin + fScale * vPlanePos;
	gluProject( vRightPos.X(), vRightPos.Y(), vRightPos.Z(), modelview, projection, viewport, & seedPos[0], &seedPos[1], &winZ );
	Wml::Vector2d d(seedPos-seedCenter);

	// add that vector to alignFrame origin (in screen space) and get ray through point
	Wml::Vector3f vOrigin2 = alignFrame.Origin();
	Wml::Vector2d seed2Center, seed2Pos;
	gluProject( vOrigin2.X(), vOrigin2.Y(), vOrigin2.Z(), modelview, projection, viewport, & seed2Center[0], &seed2Center[1], &winZ );
	seed2Pos = seed2Center + d;
	const unsigned int * pCamViewport = pCamera->GetViewport();
	Wml::Ray3f vRay;
	//! assuming that GetPickRay wants flipped Y
	pCamera->GetPickRay( (float)seed2Pos[0], (float)pCamViewport[3] - (float)seed2Pos[1], pCamViewport[2], pCamViewport[3], vRay.Origin, vRay.Direction );

	// intersect ray with alignFrame tangent plane
	Wml::ExtPlane3f plane( alignFrame.Z(), alignFrame.Origin() );
	float fT = plane.IntersectRay( vRay.Origin, vRay.Direction );
	Wml::Vector3f vHitPos = vRay.Origin + fT*vRay.Direction;

	// figure out rotation angle
	Wml::Vector3f vHitVec = vHitPos - alignFrame.Origin();
	vHitVec.Normalize();
	float x = vHitVec.Dot(alignFrame.X());
	float y = vHitVec.Dot(alignFrame.Y());
	float fAngle = atan2(y,x);
	float fWantAngle = atan2(vFrameDir.Y(), vFrameDir.X());

	return fAngle-fWantAngle;
}


void MeshObject::AlignSeedFrame2(Wml::ExtendedWmlCamera * pCamera)
{
	if (!m_bExpMap2Initialized)
		return;

	int viewport[4];
	double projection[16], modelview[16];
	glGetDoublev( GL_PROJECTION_MATRIX, projection );
	glGetDoublev( GL_MODELVIEW_MATRIX, modelview );
	glGetIntegerv( GL_VIEWPORT, viewport );

	// just use x-axis
	//Wml::Vector2f vDir(1,0);
	//float fAngle = computeAlignAngle( m_vSeedFrame, m_vSeedFrame2, vDir, viewport, projection, modelview, pCamera );

	// average 4 axes
	Wml::Vector2f vDirs[4] = { Wml::Vector2f(-1,0), Wml::Vector2f(1,0), Wml::Vector2f(0,-1), Wml::Vector2f(0,1) };
	int nDirs = 4;
	float fAngleAvg = 0;
	for ( int k = 0; k < nDirs; ++k ) {
		float fAngleK = computeAlignAngle( m_vSeedFrame, m_vSeedFrame2, vDirs[k], viewport, projection, modelview, pCamera );
		if ( fAngleK < -3.14159 )		// ugly...will this always work?? only seems to happen for (-1,0)  (depends on atan2 range)
			fAngleK += Wml::Mathf::TWO_PI;
		fAngleAvg += fAngleK;
	}
	float fAngle = fAngleAvg / (float)nDirs;

	Wml::Matrix3f matRotate;
	matRotate.FromAxisAngle(m_vSeedFrame2.Z(), fAngle);
	m_vSeedFrame2.Rotate(matRotate);

	m_bExpMapValid = false;
}


void MeshObject::ScaleExpMap( float fScale )
{
	if (!m_bExpMapInitialized)
		return;

	if ( fScale <= 0 )
		return;
	m_fDecalRadius *= fScale;
	if ( m_fDecalRadius < 0.01f )
		m_fDecalRadius = 0.01f;
	m_bExpMapValid = false;
}

void MeshObject::RotateExpMap( float fRotate )
{
	if (!m_bExpMapInitialized)
		return;

	Wml::Matrix3f matRotate;
	matRotate.FromAxisAngle(m_vSeedFrame.Z(), fRotate);
	m_vSeedFrame.Rotate(matRotate);
	m_bExpMapValid = false;
}

void MeshObject::SetSmoothNormals( bool bSet ) 
{
	m_bSmoothNormals = bSet;
	if (m_bExpMapInitialized)
		m_bExpMapValid = false;
}
void MeshObject::SetUpwindAverage( bool bSet ) 
{
	m_bUpwindAverage = bSet;
	if (m_bExpMapInitialized)
		m_bExpMapValid = false;
}
void MeshObject::SetUseConstantNormal( bool bSet )
{
	m_bUseConstantNormal = bSet;
	if (m_bExpMapInitialized)
		m_bExpMapValid = false;
}
void MeshObject::SetPreserveProjectedLengths( bool bSet )
{
	m_bPreserveProjectedLengths = bSet;
	if (m_bExpMapInitialized)
		m_bExpMapValid = false;
}




void MeshObject::ValidateExpMap()
{
	if (!m_bExpMapInitialized)
		return;
	if ( m_bExpMapValid )
		return;

	//rms::IMesh::VertexID vNearest;
	//m_bvTree.FindNearestVtx(m_vSeedFrame.Origin(), vNearest);
	//fprintf(stderr, "Nearest vertex: %d\n", vNearest);

	m_expmapgen.SetUseNeighbourNormalSmoothing( m_bSmoothNormals );
	m_expmapgen.SetUseUpwindAveraging(m_bUpwindAverage);
	m_expmapgen.SetUseConstantNormal(m_bUseConstantNormal);
	m_expmapgen.SetPreserveProjectedLengths(m_bPreserveProjectedLengths);

	float fParamRadius = m_fDecalRadius + m_fBoundaryWidth;
	m_expmapgen.SetSurfaceDistances(m_vSeedFrame.Origin(), 1.1f*m_fMaxEdgeLength, 
		fParamRadius, &m_vSeedFrame, &m_vSeedFrame.Z());
	m_expmapgen.CopyVertexUVs(&m_mesh, 0);

	if ( m_bExpMap2Initialized ) {
		m_expmapgen.SetSurfaceDistances(m_vSeedFrame2.Origin(), 1.1f*m_fMaxEdgeLength, 
			fParamRadius, &m_vSeedFrame2, &m_vSeedFrame2.Z());
		m_expmapgen.CopyVertexUVs(&m_mesh2, 0);
	}

	m_bExpMapValid = true;
}



void MeshObject::SmoothNormals(int nCount)
{
	rms::VFTriangleMesh tmp;
	tmp.Copy(m_mesh);
	for ( int k = 0; k < nCount; ++k )
		rms::MeshUtils::SmoothNormals(tmp);
	SetMesh(tmp);
}


void MeshObject::Render(bool bDrawWireframe, bool bDrawNormals)
{
	glPushAttrib(GL_ENABLE_BIT | GL_POLYGON_BIT | GL_LINE_BIT | GL_DEPTH_BUFFER_BIT );

	rms::VFMeshRenderer renderer(&m_mesh);
	//renderer.SetNormalMode( rms::VFMeshRenderer::VertexNormals );
	renderer.SetNormalMode( rms::VFMeshRenderer::FaceNormals );

	// draw wireframe
	if ( bDrawWireframe ) {
		glDisable(GL_LIGHTING);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glColor4f(0,0,1.0,1);
		renderer.Render();
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(1,1);
	}

	// render base mesh
	glEnable(GL_LIGHTING);
	glColor3f(0.6f, 1.0f, 0.6f);
	renderer.Render();

	// render support region of decal parameterization
	glDepthFunc(GL_LEQUAL);
	renderer.SetTexture2DMode( rms::VFMeshRenderer::VertexTexture2D_Required );
	glColor4f(0.5f, 0.5f, 1.0f, 1.0f);
	//renderer.Render();

	// do texture rendering

	// enable transparency and alpha test
	glDepthFunc(GL_LEQUAL);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glAlphaFunc( GL_NOTEQUAL, 0.0f );
	glEnable( GL_ALPHA_TEST );

	// enable texture
	m_texture.Enable();
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	// set texture matrix transformation so that decal is in the right spot
	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
	glTranslatef( 0.5f, 0.5f, 0.0f );	
	float fScale = 1.0f / (m_fDecalRadius * (float)sqrt(2.0f));
	glScalef( fScale, fScale, 1.0f );
	glMatrixMode(GL_MODELVIEW);

	renderer.SetTexture2DMode( rms::VFMeshRenderer::VertexTexture2D_Required );
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	renderer.Render();

	if ( m_bExpMap2Initialized ) {
		rms::VFMeshRenderer renderer2(&m_mesh2);
		renderer2.SetNormalMode( rms::VFMeshRenderer::FaceNormals );
		renderer2.SetTexture2DMode( rms::VFMeshRenderer::VertexTexture2D_Required );
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		renderer2.Render();
	}


	glMatrixMode(GL_TEXTURE);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);

	m_texture.Disable();

	glDisable(GL_LIGHTING);
	glDisable(GL_ALPHA_TEST);
	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);

	// draw seed point frames
	glLineWidth(5.0f);
	glBegin(GL_LINES);
	glColor3f(1.0f,0.0f,0.0f);
	glVertex3fv(m_vSeedFrame.Origin());
	glVertex3fv(m_vSeedFrame.Origin() + 0.05f * m_vSeedFrame.X());
	glColor3f(0.0f,1.0f,0.0f);
	glVertex3fv(m_vSeedFrame.Origin());
	glVertex3fv(m_vSeedFrame.Origin() + 0.05f * m_vSeedFrame.Y());
	glColor3f(0.0f,0.0f,1.0f);
	glVertex3fv(m_vSeedFrame.Origin());
	glVertex3fv(m_vSeedFrame.Origin() + 0.05f * m_vSeedFrame.Z());
	glEnd();

	glLineWidth(5.0f);
	glBegin(GL_LINES);
	glColor3f(1.0f,0.0f,0.0f);
	glVertex3fv(m_vSeedFrame2.Origin());
	glVertex3fv(m_vSeedFrame2.Origin() + 0.05f * m_vSeedFrame2.X());
	glColor3f(0.0f,1.0f,0.0f);
	glVertex3fv(m_vSeedFrame2.Origin());
	glVertex3fv(m_vSeedFrame2.Origin() + 0.05f * m_vSeedFrame2.Y());
	glColor3f(0.0f,0.0f,1.0f);
	glVertex3fv(m_vSeedFrame2.Origin());
	glVertex3fv(m_vSeedFrame2.Origin() + 0.05f * m_vSeedFrame2.Z());
	glEnd();


	// draw normals
	if ( bDrawNormals ) {
		glLineWidth(2.0f);
		glBegin(GL_LINES);
		size_t nCount = m_mesh.GetVertexCount();
		for ( unsigned int k = 0; k < nCount; ++k ) {
			Wml::Vector3f v,n;
			m_mesh.GetVertex(k, v, &n);
			glColor3f(0,0,0);
			glVertex3fv(v);
			glColor3f(1,0,1);
			glVertex3fv(v+0.025f*n);
		}
		glEnd();
	}

	glPopAttrib();
}



void MeshObject::RenderParamMesh()
{
	glPushAttrib(GL_ENABLE_BIT | GL_POLYGON_BIT | GL_LINE_BIT | GL_POINT_BIT);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glLineWidth(2.0f);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glColor3f(0.0f, 0.0f, 0.0f);

	// set transformation so that decal is in the right spot
	glPushMatrix();
	glLoadIdentity();
	//glTranslatef( 0.5f, 0.5f, 0.0f );	
	//float fScale = 1.0f / (m_fDecalRadius * (float)sqrt(2.0f));
	float fParamRadius = m_fDecalRadius + m_fBoundaryWidth;
	float fScale = 1.5f / (fParamRadius * (float)sqrt(2.0f));
	//fScale *= 0.2f;
	glScalef( fScale, fScale, 1.0f );

	rms::VFMeshRenderer renderer(&m_mesh);
	renderer.Render_UV();	

	glColor3f(1.0f, 0.0f, 0.0f);
	glLineWidth(3.0f);
	glBegin(GL_LINE_LOOP);
	float fSize = m_fDecalRadius / (float)sqrt(2.0f);
	glVertex2f(-fSize, -fSize);
	glVertex2f(fSize, -fSize);
	glVertex2f(fSize, fSize);
	glVertex2f(-fSize, fSize);
	glEnd();
	

	glColor3f(0.0f, 1.0f, 0.0f);
	glPointSize(5.0f);
	glBegin(GL_POINTS);
	glVertex2f(0,0);
	glEnd();

	glPopMatrix();

	glPopAttrib();
}